/* genrc - generic system initialization script helper
   Copyright (C) 2018-2025 Sergey Poznyakoff

   Genrc is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Genrc is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with genrc. If not, see <http://www.gnu.org/licenses/>. */

#include "genrc.h"

static char *
progname_pattern(char const *prog)
{
	grecs_txtacc_t acc;
	char *ret;
	static char const specials[] = "\\[]?.*+(){}|";

	acc = grecs_txtacc_create();
	grecs_txtacc_grow_char(acc, '^');
	if (prog[0] != '/')
		grecs_txtacc_grow_string(acc, "(?:/(?:.*/)?)?");
	for (; *prog; prog++) {
		if (strchr(specials, *prog))
			grecs_txtacc_grow_char(acc, '\\');
		grecs_txtacc_grow_char(acc, *prog);
	}
	grecs_txtacc_grow_char(acc, '$');
	grecs_txtacc_grow_char(acc, 0);
	ret = grecs_txtacc_finish(acc, 1);
	grecs_txtacc_free(acc);
	return ret;
}

#if HAVE_LIBPCRE == 1
# include <pcre.h>

void
match_pcre_init(PROCSCANBUF buf, char const *pattern)
{
	pcre *pcre;
	char const *error;
	int error_offset;
	char *tmp = NULL;
	int cflags = 0;
	if (buf->flags & PROCF_ICASE)
		cflags |= PCRE_CASELESS;
	if (!pattern) {
		pattern = tmp = progname_pattern(genrc_program);
	}
	pcre = pcre_compile(pattern, cflags, &error, &error_offset, 0);
	free(tmp);
	if (!pcre)
		usage_error("%s at offset %d", error, error_offset);
	buf->pattern = pcre;
}

void
match_pcre_free(PROCSCANBUF buf)
{
	pcre_free ((pcre*)buf->pattern);
}

int
match_pcre(PROCSCANBUF buf, char const *arg)
{
	int rc = pcre_exec((pcre*)buf->pattern, 0, arg, strlen(arg),
			   0, 0, NULL, 0);

	return !(rc >= 0);
	/* FIXME: error message unless rc == PCRE_ERROR_NOMATCH */
}
#elif HAVE_LIBPCRE == 2
# define PCRE2_CODE_UNIT_WIDTH 8
# include <pcre2.h>

static char *
get_error_message(int error_code)
{
	size_t errsize = 32;
	int rc;
	char *errmsg;

	errmsg = xmalloc(errsize);

	while ((rc = pcre2_get_error_message(error_code,
					     (PCRE2_UCHAR*) errmsg,
					     errsize)) == PCRE2_ERROR_NOMEMORY)
		errmsg = x2nrealloc(errmsg, &errsize, 1);
	return errmsg;
}

typedef struct {
	pcre2_code *code;
	pcre2_match_data *md;
} REGEX;

void
match_pcre_init(PROCSCANBUF buf, char const *pattern)
{
	REGEX *re;
	int flags = 0;
	char *tmp = NULL;
	int error_code;
	size_t error_offset;

	re = xcalloc(1, sizeof(*re));
	if (buf->flags & PROCF_ICASE)
		flags |= PCRE2_CASELESS;
	if (!pattern)
		pattern = tmp = progname_pattern(genrc_program);
	re->code = pcre2_compile((PCRE2_SPTR8) pattern, strlen(pattern), flags,
			     &error_code, &error_offset, NULL);
	free(tmp);
	if (!re->code)
		usage_error("%s at offset %d",
			    get_error_message(error_code), error_offset);
	re->md = pcre2_match_data_create_from_pattern(re->code, NULL);
	buf->pattern = re;
}

void
match_pcre_free(PROCSCANBUF buf)
{
	REGEX *re = buf->pattern;
	pcre2_code_free(re->code);
	pcre2_match_data_free(re->md);
	free(re);
}

int
match_pcre(PROCSCANBUF buf, char const *arg)
{
	REGEX *re = buf->pattern;
	int rc;

	rc = pcre2_match(re->code,
			 (PCRE2_SPTR8)arg, strlen(arg), 0, 0, re->md, NULL);
	if (rc > 0)
		return 0;
	switch (rc) {
	case PCRE2_ERROR_NOMATCH:
	case PCRE2_ERROR_PARTIAL:
		break;
	default:
		genrc_error("pcre2_match: %s", get_error_message(rc));
	}
	return 1;
}
#else
# error "Unrecognized value of HAVE_LIBPCRE"
#endif
